/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "initializations.h"

int settings_init(int argn,char *argv[],struct configuration_struct *settings,int sensitivity[]) {
/* that function is responsible for filling the settings structure */
/* with default values and command line arguments */
/* returns EXIT_FAILURE to terminates the program or EXIT_SUCCESS */
	char *str_extract;
	int i,j,end_loop;
	int sensitivity_default;

	/* setting default values */
	sensitivity_default=SENSITIVITY_GM45_CO60;
	for(j=0;j<SERIAL_PORTS_MAX;j++) sensitivity[j]=0;
	settings->generic_location[0]='\0';
	strncat(settings->generic_location,SERIAL_LOCATION,FILE_NAME_LENGTH);
	settings->num_probes=SERIAL_PORTS_NUMBER;
	settings->probe_index_start=SERIAL_PORTS_START;
	settings->file_descriptor_max=0;
	settings->activity_data_write_sampling=ACTIVITY_DATA_WRITE_SAMPLING;
	settings->pulse_data_write_sampling=PULSE_DATA_WRITE_SAMPLING;
	settings->pulse_mode=PULSE_MODE;
	settings->pulse_particule_sub_sampling=PULSE_PARTICULE_SUB_SAMPLING;
	settings->pulse_mode_particules_sampling_count=0;
	settings->pulse_mode_particules_sampling_write=1e3*(int)PULSE_DATA_WRITE_SAMPLING/(int)PULSE_PARTICULE_SUB_SAMPLING;
	settings->file_log=FILE_LOG;
	#ifdef SQL_LIBS
	settings->sql_log=EXIT_FAILURE;
	#endif
	
	#ifdef SQL_LIBS
	/* strings assignment */
	strncpy(settings->sql_host,SQL_HOST,SQL_PARAMETERS_STRINGS);
	settings->sql_port=SQL_PORT;
	strncpy(settings->sql_database,SQL_DATABASE,SQL_PARAMETERS_STRINGS);
	strncpy(settings->sql_table,SQL_TABLE,SQL_PARAMETERS_STRINGS);
	strncpy(settings->sql_field,SQL_FIELD,SQL_PARAMETERS_STRINGS);
	strncpy(settings->sql_user,SQL_USER,SQL_PARAMETERS_STRINGS);
	strncpy(settings->sql_password,SQL_PASSWORD,SQL_PARAMETERS_STRINGS);
	#endif
	
	/* reading command line */
	switch(argn){
		case 2:{
			if((strcmp(argv[1],"-h")==0) || (strcmp(argv[1],"--help")==0) || (strcmp(argv[1],"-help")==0)){
				fprintf(stdout,"%s %s (%s for %s, %s)\n",GM4LIN_NAME,GM4LIN_VERSION,GM4LIN_AUTHOR,GM4LIN_COMPANY,GM4LIN_DATE);
				fprintf(stdout,"recording of serial ports GM-10 and GM-45 Radiation Detectors\n");
				fprintf(stdout,"Options:\n");
				fprintf(stdout,"\t-d, --device [device root]: specify a device class (default=%s)\n",settings->generic_location);
				fprintf(stdout,"\t-n, --number [number]: select the number of probes (default=%d)\n",settings->num_probes);
				fprintf(stdout,"\t-s, --start [index probe start]: starting number for the probes (default=%d)\n",settings->probe_index_start);
				fprintf(stdout,"\t-p, --pulse: radiation pulse recording mode (default=no)\n");
				fprintf(stdout,"\t--sensitivity=[nn,nn,nn]: comma delimited list of sensitivity keywords\n");
				fprintf(stdout,"\t--default-sensitivity=[keyword]: default sensitivity keywords (default=GM45_CO60)\n");
				fprintf(stdout,"\t--list-sensitivities: gives the list of available sensitivities keywords\n");
				fprintf(stdout,"\t--no-file-log: don't log data to a file (default=log)\n");
				#ifdef SQL_LIBS
				fprintf(stdout,"\t--mysql [user]@[host]:[database]:[table]:[field] log data into a MySQL database (default=%s@%s:%s:%s:%s)\n",settings->sql_user,settings->sql_host,settings->sql_database,settings->sql_table,settings->sql_field);
				fprintf(stdout,"\t--port [port]: port used to connect to MySQL database (default=%d - MySQL default=0)\n",settings->sql_port);
				fprintf(stdout,"\t--passwd [password]: password for the MySQL database (default=%s)\n",settings->sql_password);
				#endif
				fprintf(stdout,"\t-h, --help: shows this help screen\n");
				fprintf(stdout,"\t--license: shows your rights and obligation when using this software. Please read it!\n");
				fprintf(stdout,"CRTL-C when the software is running stops recording activity and exits\n");
				fprintf(stdout,"------------------\n");
				fprintf(stdout,"%s\n",GM4LIN_AUTHOR_EMAIL);
				return EXIT_FAILURE;
			}
		}
	}
	end_loop=argn-1;
	/* command line arguments scan */
	for(i=1;i<=(end_loop);i++){
		if((strcasecmp(argv[i],"--license")==0)) {
			fprintf(stdout,"This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.\n");
			fprintf(stdout,"This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License in file \"COPYING\" for more details, or write to\n");
			fprintf(stdout,"\tFree Software Foundation, Inc. 675 Mass Ave, Cambridge, MA 02139, USA\n");
			fprintf(stdout,"to obtain a copy of the license.\n");
			return EXIT_FAILURE;
		}
		if((strcasecmp(argv[i],"-d")==0) || (strcasecmp(argv[i],"--device")==0)) {
			/* selecting a device */
			if(++i>end_loop) {
				fprintf(std_err,"settings_init: not enought parameters for device location!\n");
				return EXIT_FAILURE;
			}
			sscanf(argv[i],"%s",settings->generic_location);
			continue;
		}
		if((strcasecmp(argv[i],"-n")==0) || (strcasecmp(argv[i],"--number")==0)) {
			/* selecting the number of probes */
			if(++i>end_loop) {
				fprintf(std_err,"settings_init: not enought parameters for number of devices!\n");
				return EXIT_FAILURE;
			}
			sscanf(argv[i],"%d",&settings->num_probes);
			if (settings->num_probes>SERIAL_PORTS_MAX) {
				perror("settings_init: too many probes required\n");
				return EXIT_FAILURE;	
			}
			continue;
		}
		if((strcasecmp(argv[i],"-s")==0) || (strcasecmp(argv[i],"--start")==0)){
			/* selecting the index of the first probe */
			if(++i>end_loop) {
				fprintf(std_err,"settings_init: not enought parameters for first device number!\n");
				return EXIT_FAILURE;
			}
			sscanf(argv[i],"%d",&settings->probe_index_start);
			continue;
		}
		if((strcasecmp(argv[i],"-p")==0) || (strcasecmp(argv[i],"--pulse")==0)){
			/* switching to pulse recording mode */
			settings->pulse_mode=EXIT_SUCCESS;
			continue;
		}
		if(strncasecmp(argv[i],"--sensitivity",strlen("--sensitivity"))==0){
			/* reading the list of index for gm10 detectors */
			j=0;
			str_extract=(char *)strtok(argv[i],"=");
			str_extract=(char *)strtok(NULL,"=");
			str_extract=(char *)strtok(str_extract,",");
			while((str_extract!=NULL) && (j<SERIAL_PORTS_MAX)) {
				if(strcasecmp(str_extract,"GM45_CO60")==0) sensitivity[j]=SENSITIVITY_GM45_CO60;
				if(strcasecmp(str_extract,"GM10_CO60")==0) sensitivity[j]=SENSITIVITY_GM10_CO60;
				if(strcasecmp(str_extract,"GM10_CS137")==0) sensitivity[j]=SENSITIVITY_GM10_CS137;
				str_extract=(char *)strtok(NULL,",");
				j++;
			}
			if(j>=SERIAL_PORTS_MAX) perror("initialization.c: too many sensitivities given, ignoring them\n");
			continue;
		}
		if(strncasecmp(argv[i],"--default-sensitivity",strlen("--default-sensitivity"))==0){
			/* reading and writing default sensitivity in the sensitivity table */
			str_extract=(char *)strtok(argv[i],"=");
			str_extract=(char *)strtok(NULL,"=");
			if(str_extract!=NULL) {
				if(strcasecmp(str_extract,"GM45_CO60")==0) sensitivity_default=SENSITIVITY_GM45_CO60;
				if(strcasecmp(str_extract,"GM10_CO60")==0) sensitivity_default=SENSITIVITY_GM10_CO60;
				if(strcasecmp(str_extract,"GM10_CS137")==0) sensitivity[j++]=SENSITIVITY_GM10_CS137;
			} else {
				perror("settings_init: wrong syntax for default sensitivity selection\n");
				return EXIT_FAILURE;
			}
			continue;
		}
		if(strcasecmp(argv[i],"--list-sensitivities")==0){
			/* giving the list of available sensitivities */
			fprintf(std_err,"Available sensitivity keywords and their values (in cps/mR/h):\n");
			fprintf(std_err,"GM45_CO60 -> %d\n",SENSITIVITY_GM45_CO60);
			fprintf(std_err,"GM10_CO60 -> %d\n",SENSITIVITY_GM10_CO60);
			fprintf(std_err,"GM10_CS137 -> %d\n",SENSITIVITY_GM10_CS137);
			return EXIT_FAILURE;
		}
		if(strcasecmp(argv[i],"--no-file-log")==0){
			/* switching to pulse recording mode */
			settings->file_log=EXIT_FAILURE;
			continue;
		}
		#ifdef SQL_LIBS
		if((strcasecmp(argv[i],"--mysql")==0)){
			/* switching to MySQL recording mode */
			settings->sql_log=EXIT_SUCCESS;
			if(++i<=end_loop) {
				/* look if the next argument goes with --mysql or is another option */
				if(argv[i][1]=='-') {
					i--;
					continue;
				}
				/* now we can read the parameters of the sql connection */
				str_extract=(char *)strtok(argv[i],"@");
				if(str_extract!=NULL) strncpy(settings->sql_user,str_extract,SQL_PARAMETERS_STRINGS);
				str_extract=(char *)strtok(NULL,":");
				if(str_extract!=NULL) strncpy(settings->sql_host,str_extract,SQL_PARAMETERS_STRINGS);
				str_extract=(char *)strtok(NULL,":");
				if(str_extract!=NULL) strncpy(settings->sql_database,str_extract,SQL_PARAMETERS_STRINGS);
				str_extract=(char *)strtok(NULL,":");
				if(str_extract!=NULL) strncpy(settings->sql_table,str_extract,SQL_PARAMETERS_STRINGS);
				str_extract=(char *)strtok(NULL,":");
				if(str_extract!=NULL) strncpy(settings->sql_field,str_extract,SQL_PARAMETERS_STRINGS);
			}
			continue;
		}
		if((strcasecmp(argv[i],"--port")==0)){
			/* port for MySQL recording mode */
			if(++i>end_loop) {
				fprintf(std_err,"settings_init: not enought parameters for port selection!\n");
				return EXIT_FAILURE;
			}
			sscanf(argv[i],"%d",&settings->sql_port);
			continue;
		}
		if((strcasecmp(argv[i],"--passwd")==0)){
			/* password for MySQL recording mode */
			if(++i>end_loop) {
				fprintf(std_err,"settings_init: not enought parameters for mysql password!\n");
				return EXIT_FAILURE;
			}
			sscanf(argv[i],"%s",settings->sql_password);
			continue;
		}
		#endif
		
	}
	/* replacing in the table the default sensitivity by its value */
	for(j=0;j<SERIAL_PORTS_MAX;j++) {
		if(sensitivity[j]==0) sensitivity[j]=sensitivity_default;
	}
	
	return EXIT_SUCCESS;

}

int probes_init(struct probe_descriptor probe[], struct configuration_struct *settings,int sensitivity[]) {
/* initialization of the probes */
/* return EXIT_SUCCESS or EXIT_FAILURE */
	int probe_index,n;
	
	for(probe_index=0;probe_index<settings->num_probes;probe_index++) {
		/* sensitivity of the probe */
		probe[probe_index].sensitivity=sensitivity[probe_index];
		/* conversion factor from signal to R/h */
		if (settings->pulse_mode==EXIT_SUCCESS) {
			probe[probe_index].conv2rh=1.0/(bicron_signal2Rh*probe[probe_index].sensitivity);
			probe[probe_index].conv2display=1.0;
		} else {
			probe[probe_index].conv2rh=1.0/(events2Rh*probe[probe_index].sensitivity);
			probe[probe_index].conv2display=events2cpm;
		}
		/* units used for display and real science */
		if (settings->pulse_mode==EXIT_SUCCESS) {
			n=snprintf(probe[probe_index].display_units,SHORT_STRING,"raw");
			if (n <= -1 || n >= (int)(SHORT_STRING*sizeof(char))) {
				perror("probe_init: display_unit string building failure\n");
				return EXIT_FAILURE;
			}
			n=snprintf(probe[probe_index].si_units,SHORT_STRING,"R/h");
			if (n <= -1 || n >= (int)(SHORT_STRING*sizeof(char))) {
				perror("probe_init: si_unit string building failure\n");
				return EXIT_FAILURE;
			}
		} else {
			n=snprintf(probe[probe_index].display_units,SHORT_STRING,"cpm");
			if (n <= -1 || n >= (int)(SHORT_STRING*sizeof(char))) {
				perror("probe_init: display_unit string building failure\n");
				return EXIT_FAILURE;
			}
			n=snprintf(probe[probe_index].si_units,SHORT_STRING,"R/h");
			if (n <= -1 || n >= (int)(SHORT_STRING*sizeof(char))) {
				perror("probe_init: si_unit string building failure\n");
				return EXIT_FAILURE;
			}
		}
		
		/* initializing and run the probes */
		if((start_probe(&probe[probe_index],probe_index+settings->probe_index_start,settings)) == EXIT_FAILURE) {
			perror("probe_init: probe can not be started");
			/* closing the log file */
			if (settings->file_log==EXIT_SUCCESS) {
				if ((fclose(probe[probe_index].log_file)) != EXIT_SUCCESS){
					perror("probe_init: Log file can not be closed");
				}
			}
			return EXIT_FAILURE;
		}
		if((probe[probe_index].file_descriptor) > (settings->file_descriptor_max)) settings->file_descriptor_max=probe[probe_index].file_descriptor;
		
		/* giving some information about what is going on */
		printf("\tprobe[%d].sensitivity=%d\n",probe_index+1,probe[probe_index].sensitivity);
		if (settings->file_log==EXIT_SUCCESS) {
			fprintf(probe[probe_index].log_file,"#Probe[%d] sensitivity factor=%d\n",probe_index+1,probe[probe_index].sensitivity);
		}
	}
	return EXIT_SUCCESS;
}

#ifdef SQL_LIBS
int sql_init(struct configuration_struct *settings) {
/* initialization of the sql database connection */
/* return EXIT_SUCCESS or EXIT_FAILURE */
	if(mysql_init(&settings->mysql)==NULL) {
		fprintf(stderr,"sql_init: Failed to initialize MySQL connection\n");
		return EXIT_FAILURE;
	}
	if (!mysql_real_connect(&settings->mysql,settings->sql_host,settings->sql_user,settings->sql_password,settings->sql_database,settings->sql_port,NULL,0)) {
		fprintf(stderr,"sql_init: Failed to connect to database: Error: %s\n", mysql_error(&settings->mysql));
		return EXIT_FAILURE;
	}
	if (mysql_select_db(&settings->mysql,settings->sql_database)) {
		fprintf(stderr,"sql_init: Couldn't select database %s!\n%s\n",settings->sql_database, mysql_error(&settings->mysql));
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}
#endif

void sig_alrm(void) {
/* function called when a SIGALRM signal is catched */
	flag_alrm=1;
}

void sig_stop(void) {
/* function called when a SIGINT signal is catched */
	fprintf(std_msg,"************\nProgram stopping...\n");
	flag_stop=1;
}

int stop_catch_init(void) {
/* this is the initialization of signal catching for the STOP signal */
/* we want to catch the signal SIGINT to stop the program in a clean way*/
/* return EXIT_SUCCES or EXIT_FAILURE */
	struct sigaction stop_act;			/* for signal catching (for the end of recording) */
	
    	stop_act.sa_handler=(void *)sig_stop;
    	/* We don't want to block any other signals */
    	sigemptyset(&stop_act.sa_mask);
	if (sigaction(SIGINT,&stop_act,NULL) != EXIT_SUCCESS) {
		perror("stop_catch_init: signal catching failed for SIGSTOP");
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}

int timer_init(long write_timer_duration, long short_timer_duration) {
/* this is the initialization of signal catching and timer */
/* first argument is the timer duration in seconds, the second one is in ms */
/* return EXIT_SUCCES or EXIT_FAILURE */
	struct itimerval timer_parameters;		/* parameters for the timer, in order to calculate cpm, dose... */
	struct sigaction timer_act;			/* for signal catching (for the timer) */
	
	/* we want to catch the signal SIGALRM in order to use it as a timer */
    	timer_act.sa_handler=(void *)sig_alrm;
    	/* We don't want to block any other signals */
    	sigemptyset(&timer_act.sa_mask);
	if (sigaction(SIGALRM,&timer_act,NULL) != EXIT_SUCCESS) {
		perror("timer_init: signal catching failed for SIGALRM");
		return EXIT_FAILURE;
	}
	
	/* setting up a timer, which is going to send SIGALRM every average_length seconds */
	timer_parameters.it_value.tv_sec=write_timer_duration;
	timer_parameters.it_value.tv_usec=short_timer_duration;
	timer_parameters.it_interval.tv_sec=write_timer_duration;
	timer_parameters.it_interval.tv_usec=short_timer_duration;
	flag_alrm=FALSE;
	if((setitimer (ITIMER_REAL,&timer_parameters,NULL)) != EXIT_SUCCESS) {
		perror("timer_init: timer setup failed");
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}

int timer_sync(void) {
/* wait until the number of seconds is '00' */
/* return EXIT_SUCCES or EXIT_FAILURE */
	time_t timing;				/* time in seconds since EPOC */
	
	fprintf(std_msg,"\tWaiting for seconds to be \"00\"\n");
	do {
		if(time(&timing) == ((time_t)-1)) {
			perror("timer_sync: Time reading failure\n");
			return EXIT_FAILURE;
		}
	} while(((int)timing % 60) != 0);
	
	return EXIT_SUCCESS;
}

int sequencer_init(struct configuration_struct settings) {
/* starting signal catching and timer */
/* return EXIT_SUCCESS or EXIT_FAILURE */
	struct timeval timeout;			/* time out structure for "select" */
	if(stop_catch_init() == EXIT_FAILURE) {
		perror("sequencer_init: SIGINT catching initialization failure");
		return EXIT_FAILURE;
	}
	//timer_sync();
	if(settings.pulse_mode==EXIT_FAILURE) {
		if(timer_init(settings.activity_data_write_sampling,0) == EXIT_FAILURE) {
			perror("sequencer_init: Timer initialization failure");
			return EXIT_FAILURE;
		}
	} else {
		if(timer_init(0,settings.pulse_particule_sub_sampling*1e3) == EXIT_FAILURE) {
			perror("sequencer_init: Timer initialization failure");
			return EXIT_FAILURE;
		}
		settings.pulse_mode_particules_sampling_write=1e3*settings.pulse_data_write_sampling/settings.pulse_particule_sub_sampling;
	}
	return EXIT_SUCCESS;
}
